import abc
import asyncio
import functools
import json
from asyncio.events import AbstractEventLoop
from asyncio.exceptions import CancelledError
from asyncio.locks import Event
from asyncio.tasks import Task
from typing import Callable, Dict, Optional, Set, TypeVar, Union

import aiohttp.web_exceptions
from aiohttp import ClientSession, FormData
from aiohttp.client_ws import ClientWebSocketResponse
from aiohttp.http_websocket import WSMsgType
from graia.broadcast import Broadcast
from graia.broadcast.utilles import run_always_await
from loguru import logger
from typing_extensions import ParamSpec
from yarl import URL

from graia.ariadne.event import MiraiEvent
from graia.ariadne.event.network import RemoteException
from graia.ariadne.exception import InvalidArgument, InvalidSession, NotSupportedAction
from graia.ariadne.model import CallMethod, MiraiSession
from graia.ariadne.util import validate_response

P = ParamSpec("P")
R = TypeVar("R")


def require_verified(func: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(func)
    def wrapper(self: "Adapter", *args: P.args, **kwargs: P.kwargs):
        if not self.mirai_session.session_key:
            raise InvalidSession("you must verify the session before action.")
        return func(self, *args, **kwargs)

    return wrapper


def error_wrapper(network_action_callable: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(network_action_callable)
    async def wrapped_network_action_callable(
        self: "Adapter", *args: P.args, **kwargs: P.kwargs
    ):
        running_count = 0

        while running_count < 5:
            running_count += 1
            try:
                return await network_action_callable(self, *args, **kwargs)
            except InvalidSession as invalid_session_exc:
                logger.error(
                    "Invalid session detected, asking daemon to restart adapter..."
                )
                logger.exception(invalid_session_exc)
                await self.stop()
            except aiohttp.web_exceptions.HTTPNotFound:
                raise NotSupportedAction(
                    f"{network_action_callable.__name__}: this action not supported"
                )
            except aiohttp.web_exceptions.HTTPInternalServerError as e:
                self.broadcast.postEvent(RemoteException())
                logger.error(
                    "An exception has thrown by remote, please check the console!"
                )
                raise
            except (
                aiohttp.web_exceptions.HTTPMethodNotAllowed,
                aiohttp.web_exceptions.HTTPRequestURITooLong,
                aiohttp.web_exceptions.HTTPTooManyRequests,
            ):

                logger.error(
                    f"It seems that we post in a wrong way "
                    f"for the action '{network_action_callable.__name__}', please open a issue."
                )
                raise
            except aiohttp.web_exceptions.HTTPRequestTimeout:
                logger.error(
                    f"timeout on {network_action_callable.__name__}, retry after 5 seconds...".format()
                )
                await asyncio.sleep(5)
                continue

    return wrapped_network_action_callable


class Adapter(abc.ABC):
    """
    适配器抽象基类。

    Args:
        broadcast(Broadcast): Broadcast 实例
        session: Session 实例，存储了连接信息
    """

    def __init__(self, broadcast: Broadcast, mirai_session: MiraiSession) -> None:
        self.broadcast = broadcast
        self.loop: AbstractEventLoop = broadcast.loop
        self.mirai_session: MiraiSession = mirai_session
        self.session: Optional[ClientSession] = None
        self.running: bool = False
        self.fetch_task: Optional[Task] = None

    @abc.abstractmethod
    async def fetch_cycle(self) -> None:
        """
        负责接收并处理数据，广播事件。
        """

    @abc.abstractmethod
    @require_verified
    @error_wrapper
    async def call_api(
        self, action: str, method: CallMethod, data: Optional[dict] = None
    ) -> Union[dict, list]:
        """
        向Mirai端发送数据。
        如有回复则应一并返回。

        Args:
            action (str): 要执行的操作。
            method (CallMethod): 指示对 mirai-api-http 端发送数据的方式。
            data (dict): 要发送的数据。
        Returns:
            dict: 响应字典。
        """

    async def build_event(self, data: dict) -> MiraiEvent:
        """
        从尚未明确指定事件类型的对象中获取事件的定义, 并进行解析

        Args:
            data (dict): 用 dict 表示的序列化态事件, 应包含有字段 `type` 以供分析事件定义.

        Raises:
            InvalidArgument: 目标对象中不包含字段 `type`
            ValueError: 没有找到对应的字段, 通常的, 这意味着应用获取到了一个尚未被定义的事件, 请报告问题.

        Returns:
            MiraiEvent: 已经被序列化的事件
        """
        event_type: Optional[str] = data.get("type")
        if not event_type or not isinstance(event_type, str):
            raise InvalidArgument("Unable to find 'type' field for automatic parsing")
        event_class: Optional[MiraiEvent] = self.broadcast.findEvent(event_type)
        if not event_class:
            logger.error(
                "An event is not recognized! Please report with your log to help us diagnose."
            )
            raise ValueError(f"Unable to find event: {event_type}", data)
        data = {k: v for k, v in data.items() if k != "type"}
        obj = event_class.parse_obj(data)
        return await run_always_await(obj)

    async def start(self):
        if not self.session:
            self.session = ClientSession()
        if not self.fetch_task or self.fetch_task.done():
            self.running = True
            self.fetch_task = self.loop.create_task(self.fetch_cycle())

    @property
    def session_activated(self) -> bool:
        return bool(self.mirai_session.session_key)

    async def stop(self):
        """
        停止适配器，并等待 `fetch_cycle` 方法完成。
        """
        self.running = False
        if self.fetch_task and not self.fetch_task.done():
            try:
                await self.fetch_task
            except CancelledError:
                pass
        self.mirai_session.session_key = None


class HttpAdapter(Adapter):
    """
    仅使用正向 HTTP 的适配器，采用短轮询接收事件/消息。
    不推荐。
    Note: Working In Progress
    """

    def __init__(
        self,
        broadcast: Broadcast,
        mirai_session: MiraiSession,
        fetch_interval: float = 0.5,
    ) -> None:
        super.__init__(self, broadcast, mirai_session)
        self.fetch_interval = fetch_interval
        raise NotImplementedError("HTTP Adapter is not supported yet!")

    async def fetch_cycle(self):
        while self.running:
            await asyncio.sleep(self.fetch_interval)

    @require_verified
    @error_wrapper
    async def call_api(
        self, action: str, method: CallMethod, data: Optional[dict] = {}
    ) -> Union[dict, list]:
        data = data or dict()
        if method == CallMethod.GET or method == CallMethod.RESTGET:
            async with self.session.get(
                URL(self.mirai_session.url_gen(action)).with_query(data)
            ) as response:
                response.raise_for_status()
                resp_json: dict = await response.json()
        elif method == CallMethod.POST or method == CallMethod.RESTPOST:
            async with self.session.post(
                self.mirai_session.url_gen(action), data=json.dumps(data)
            ) as response:
                response.raise_for_status()
                resp_json: dict = await response.json()
        else:  # MULTIPART
            form = FormData()
            for k, v in data:
                form.add_fields(k, v)
            async with self.session.post(
                self.mirai_session.url_gen(action), data=form
            ) as response:
                response.raise_for_status()
                resp_json: dict = await response.json()
        validate_response(resp_json)
        if "data" in resp_json:
            return resp_json["data"]
        return resp_json


class WebsocketAdapter(Adapter):
    """
    仅使用正向 Websocket 的适配器。
    因 Mirai API HTTP 的实现，部分功能不可用。
    """

    class SyncIdManager:
        allocated: Set[int] = {0}

        @classmethod
        def allocate(cls) -> int:
            new_id = max(cls.allocated) + 1
            cls.allocated.add(new_id)
            return new_id

        @classmethod
        def done(cls, sync_id: int) -> None:
            if sync_id in cls.allocated:
                cls.allocated.remove(sync_id)

    class CallResponse(Event):
        def __init__(self, *, loop: AbstractEventLoop = None) -> None:
            super().__init__(loop=loop)
            self.response: Optional[dict] = None

    def __init__(
        self, bcc: Broadcast, mirai_session: MiraiSession, ping: bool = True
    ) -> None:
        super().__init__(bcc, mirai_session)
        self.ping = ping
        self.ping_task: Optional[Task] = None
        self.ws_conn: Optional[ClientWebSocketResponse] = None
        self.query_dict = {"verifyKey": mirai_session.verify_key}
        self.sync_event: Dict[int, WebsocketAdapter.CallResponse] = {}
        if not mirai_session.single_mode:
            self.query_dict["qq"] = mirai_session.account

    async def ws_ping(self, interval: float = 30.0) -> None:
        while self.running:
            try:
                try:
                    await self.ws_conn.ping()
                    logger.debug("websocket: ping")
                except:
                    logger.exception("websocket: ping failed")
                else:
                    logger.debug(f"websocket: ping success, delay {interval}s")
                    await asyncio.sleep(interval)
            except asyncio.CancelledError:
                logger.debug("websocket: pinger exit")
                return

    @require_verified
    @error_wrapper
    async def call_api(
        self, action: str, method: CallMethod, data: Optional[dict] = None
    ) -> Union[dict, list]:
        data = data or dict()
        if not self.ws_conn:
            raise ValueError("no existing websocket connection")
        sync_id = self.SyncIdManager.allocate()
        event = self.CallResponse()
        self.sync_event[sync_id] = event
        content = {
            "syncId": sync_id,
            "command": action,
            "content": json.dumps(data),
        }
        if method == CallMethod.RESTGET:
            content["subCommand"] = "get"
        elif method == CallMethod.RESTPOST:
            content["subCommand"] = "update"
        elif method == CallMethod.MULTIPART:
            raise NotImplementedError(
                f"Unsupported operation for WebsocketAdapter: {method}"
            )

        await self.ws_conn.send_json(content)
        logger.debug(f"websocket：sent with sync id: {sync_id}")
        await event.wait()
        self.SyncIdManager.done(sync_id)
        value: dict = event.response
        del self.sync_event[sync_id]
        del event
        validate_response(value)
        if "data" in value:
            return value["data"]
        else:
            return value

    async def raw_data_parser(self, raw_data: dict) -> None:
        sync_id: int = raw_data["syncId"]
        received_data: dict = raw_data["data"]
        validate_response(received_data)
        if not self.mirai_session.session_key:
            if session_key := received_data.get("session", None):
                self.mirai_session.session_key = session_key
            return
        if sync_id not in self.SyncIdManager.allocated:
            event = await self.build_event(received_data)
            self.broadcast.postEvent(event)
        else:
            if sync_id in self.sync_event:
                response = self.sync_event[sync_id]
                response.response = received_data
                response.set()

    async def fetch_cycle(self) -> None:
        async with self.session.ws_connect(
            str(URL(self.mirai_session.url_gen("all")).with_query(self.query_dict)),
            autoping=False,
        ) as connection:
            logger.info("websocket: connected")
            self.ws_conn = connection

            if self.ping:
                self.ping_task = self.loop.create_task(self.ws_ping())
                logger.info("websocket: ping task created")
            try:
                while self.running:
                    ws_message = await connection.receive()
                    if ws_message.type is WSMsgType.TEXT:
                        original_data: dict = json.loads(ws_message.data)
                        await self.raw_data_parser(original_data)

                    elif ws_message.type is WSMsgType.CLOSED:
                        logger.info("websocket: connection has been closed.")
                        return
                    elif ws_message.type is WSMsgType.PONG:
                        logger.debug("websocket: received pong")
                    else:
                        logger.debug(
                            "websocket: unknown message type - {}".format(
                                ws_message.type
                            )
                        )
            finally:
                if self.ping_task:
                    self.ping_task.cancel()
                    self.ping_task = None
                    logger.debug("websocket: ping task complete")
        logger.info("websocket: disconnected")


class CombinedAdapter(Adapter):
    """
    使用正向Websocket接收事件与消息，用HTTP发送消息/操作的适配器。
    稳定与性能的平衡，但需要 Mirai API HTTP 同时启用 `http` 与 `ws` 适配器。

    Args:
        bcc(Broadcast): Broadcast 实例
        session: Session 实例，存储了连接信息
        ping(bool): 是否启用 ping 功能。
    """

    def __init__(
        self, bcc: Broadcast, mirai_session: MiraiSession, ping: bool = True
    ) -> None:
        super().__init__(bcc, mirai_session)
        self.ping = ping
        self.ping_task: Optional[Task] = None
        self.ws_conn: Optional[ClientWebSocketResponse] = None
        self.query_dict = {"verifyKey": mirai_session.verify_key}
        if not mirai_session.single_mode:
            self.query_dict["qq"] = mirai_session.account

    ws_ping = WebsocketAdapter.ws_ping

    call_api = HttpAdapter.call_api

    async def raw_data_parser(self, raw_data: dict) -> None:
        received_data = raw_data["data"]
        validate_response(received_data)
        if not self.mirai_session.session_key:
            if session_key := received_data.get("session", None):
                self.mirai_session.session_key = session_key
            return
        event = await self.build_event(received_data)
        self.broadcast.postEvent(event)

    fetch_cycle = WebsocketAdapter.fetch_cycle


DefaultAdapter = CombinedAdapter


class DebugAdapter(DefaultAdapter):
    """
    Debugging adapter
    """

    async def build_event(self, data: dict) -> MiraiEvent:
        """
        从尚未明确指定事件类型的对象中获取事件的定义, 并进行解析

        Args:
            data (dict): 用 dict 表示的序列化态事件, 应包含有字段 `type` 以供分析事件定义.

        Raises:
            InvalidArgument: 目标对象中不包含字段 `type`
            ValueError: 没有找到对应的字段, 通常的, 这意味着应用获取到了一个尚未被定义的事件, 请报告问题.

        Returns:
            MiraiEvent: 已经被序列化的事件
        """
        try:
            event = await super().build_event(data)
        except ValueError as e:
            logger.error(f"{e.args[0]}\n{json.dumps(data, indent=4)}")
            raise
        else:
            logger.debug(event)
            return event



from datetime import datetime
from typing import List, Union

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from graia.ariadne.dispatcher import (
    ApplicationDispatcher,
    MessageChainDispatcher,
    SourceDispatcher,
)
from graia.ariadne.message.chain import MessageChain
from graia.ariadne.model import Client, Friend, Group, Member

from . import MiraiEvent


class MessageEvent(MiraiEvent):
    type: str = "MessageEvent"
    messageChain: MessageChain
    sender: Union[Friend, Member, Client]

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainDispatcher, ApplicationDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface):
            pass


class FriendMessage(MessageEvent):
    type: str = "FriendMessage"
    messageChain: MessageChain
    sender: Friend

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainDispatcher, ApplicationDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["FriendMessage"]):
            if interface.annotation is Friend:
                return interface.event.sender


class GroupMessage(MessageEvent):
    type: str = "GroupMessage"
    messageChain: MessageChain
    sender: Member

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainDispatcher, ApplicationDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["GroupMessage"]):
            if interface.annotation is Group:
                return interface.event.sender.group
            elif interface.annotation is Member:
                return interface.event.sender


class TempMessage(MessageEvent):
    type: str = "TempMessage"
    messageChain: MessageChain
    sender: Member

    @classmethod
    def parse_obj(cls, obj):
        return super().parse_obj(obj)

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainDispatcher, ApplicationDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["TempMessage"]):
            if interface.annotation is Group:
                return interface.event.sender.group
            elif interface.annotation is Member:
                return interface.event.sender


class OtherClientMessage(MessageEvent):
    type: str = "OtherClientMessage"
    messageChain: MessageChain
    sender: Client

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainDispatcher, ApplicationDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["OtherClientMessage"]):
            if interface.annotation is Client:
                return interface.event.sender


class StrangerMessage(MessageEvent):
    type: str = "StrangerMessage"
    messageChain: MessageChain
    sender: Friend  # use Friend because it has the same structure as the stranger.

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainDispatcher, ApplicationDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["StrangerMessage"]):
            if interface.annotation is Friend:
                return interface.event.sender



from typing import TYPE_CHECKING

if not TYPE_CHECKING:  # for init event and message
    import graia.ariadne.event.lifecycle
    import graia.ariadne.event.message
    import graia.ariadne.event.mirai
    import graia.ariadne.event.network
    import graia.ariadne.message.chain
    import graia.ariadne.message.element



import typing

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

if typing.TYPE_CHECKING:
    from graia.ariadne.app import Ariadne


class ApplicationLaunched(Dispatchable):
    app: "Ariadne"

    def __init__(self, app) -> None:
        self.app = app

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: "DispatcherInterface[ApplicationLaunched]"):
            from .. import GraiaMiraiApplication

            if interface.annotation is GraiaMiraiApplication:
                return interface.event.app


class ApplicationShutdowned(Dispatchable):
    app: "Ariadne"

    def __init__(self, app) -> None:
        self.app = app

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: "DispatcherInterface[ApplicationShutdowned]"):
            from .. import GraiaMiraiApplication

            if interface.annotation is GraiaMiraiApplication:
                return interface.event.app



import abc
import sys
from base64 import b64encode
from datetime import datetime
from enum import Enum
from json import dumps as j_dump
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

import aiohttp
from aiohttp import ClientSession
from pydantic import BaseModel, validator
from pydantic.fields import Field

from graia.ariadne.context import application_ctx, upload_method_ctx
from graia.ariadne.exception import InvalidArgument
from graia.ariadne.model import AriadneBaseModel, UploadMethod

if TYPE_CHECKING:
    from graia.ariadne.message.chain import MessageChain


class Element(AriadneBaseModel, abc.ABC):
    """
    指示一个消息中的元素。
    type (str): 元素类型
    """

    type: str

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

    def asDisplay(self) -> str:
        return ""

    def prepare(self) -> None:
        """
        为元素被发送进行准备
        若无异常被引发，则完成本方法后元素应可被发送。
        保留空实现以允许不需要 `prepare`的元素类型存在。
        """


class Plain(Element):
    type: str = "Plain"
    text: str

    def __init__(self, text: str, **kwargs) -> None:
        """实例化一个 Plain 消息元素, 用于承载消息中的文字.

        Args:
            text (str): 元素所包含的文字
        """
        super().__init__(text=text, **kwargs)

    def asDisplay(self) -> str:
        return self.text


class Source(Element):
    "表示消息在一个特定聊天区域内的唯一标识"
    type: str = "Source"
    id: int
    time: datetime

    class Config:
        json_encoders = {
            datetime: lambda v: int(v.timestamp()),
        }


class Quote(Element):
    "表示消息中回复其他消息/用户的部分, 通常包含一个完整的消息链(`origin` 属性)"
    type: str = "Quote"
    id: int
    groupId: int
    senderId: int
    targetId: int
    origin: "MessageChain"

    @validator("origin", pre=True, allow_reuse=True)
    def _(cls, v):
        from .chain import MessageChain

        return MessageChain(v)  # no need to parse objects, they are universal!


class At(Element):
    """该消息元素用于承载消息中用于提醒/呼唤特定用户的部分."""

    type: str = "At"
    target: int
    display: Optional[str] = None

    def __init__(self, target: int, **kwargs) -> None:
        """实例化一个 At 消息元素, 用于承载消息中用于提醒/呼唤特定用户的部分.

        Args:
            target (int): 需要提醒/呼唤的特定用户的 QQ 号(或者说 id.)
        """
        super().__init__(target=target, **kwargs)

    def prepare(self) -> None:
        try:
            if upload_method_ctx.get() != UploadMethod.Group:
                raise InvalidArgument(
                    "you cannot use this element in this method: {0}".format(
                        upload_method_ctx.get().value
                    )
                )
        except LookupError:
            pass

    def asDisplay(self) -> str:
        return f"[At:{self.display}({self.target})]"


class AtAll(Element):
    "该消息元素用于群组中的管理员提醒群组中的所有成员"
    type: str = "AtAll"

    def asDisplay(self) -> str:
        return "[AtAll]"

    def prepare(self) -> None:
        try:
            if upload_method_ctx.get() != UploadMethod.Group:
                raise InvalidArgument(
                    "you cannot use this element in this method: {0}".format(
                        upload_method_ctx.get().value
                    )
                )
        except LookupError:
            pass


class Face(Element):
    "表示消息中所附带的表情, 这些表情大多都是聊天工具内置的."
    type: str = "Face"
    faceId: int
    name: Optional[str] = None

    def asDisplay(self) -> str:
        return f"[表情:{self.faceId}]"


class Xml(Element):
    type = "Xml"
    xml: str

    def asDisplay(self) -> str:
        return "[XML消息]"


class Json(Element):
    type = "Json"
    Json: str = Field(..., alias="json")

    def __init__(self, json: Union[dict, str], **kwargs) -> None:
        if isinstance(json, dict):
            json = j_dump(json)
        super().__init__(json=json, **kwargs)

    def dict(self, *args, **kwargs):
        return super().dict(*args, **({**kwargs, "by_alias": True}))

    def asDisplay(self) -> str:
        return "[JSON消息]"


class App(Element):
    type = "App"
    content: str

    def asDisplay(self) -> str:
        return "[APP消息]"


class PokeMethods(Enum):
    ChuoYiChuo = "ChuoYiChuo"
    BiXin = "BiXin"
    DianZan = "DianZan"
    XinSui = "XinSui"
    LiuLiuLiu = "LiuLiuLiu"
    FangDaZhao = "FangDaZhao"
    BaoBeiQiu = "BaoBeiQiu"
    Rose = "Rose"
    ZhaoHuanShu = "ZhaoHuanShu"
    RangNiPi = "RangNiPi"
    JeiYin = "JeiYin"
    ShouLei = "ShouLei"
    GouYin = "GouYin"
    ZhuaYiXia = "ZhuaYiXia"
    SuiPing = "SuiPing"
    QiaoMen = "QiaoMen"


class Poke(Element):
    type = "Poke"
    name: PokeMethods

    def asDisplay(self) -> str:
        return f"[戳一戳:{self.name}]"


class Dice(Element):
    type = "Dice"
    value: int

    def asDisplay(self) -> str:
        return f"[骰子:{self.value}]"


class MusicShare(Element):
    type = "MusicShare"
    kind: Optional[str]
    title: Optional[str]
    summary: Optional[str]
    jumpUrl: Optional[str]
    pictureUrl: Optional[str]
    musicUrl: Optional[str]
    brief: Optional[str]

    def asDisplay(self) -> str:
        return f"[音乐分享:{self.title}]"


class ForwardNode(BaseModel):
    senderId: int
    time: datetime
    senderName: str
    messageChain: Optional["MessageChain"]
    messageId: Optional[int]

    class Config:
        json_encoders = {
            datetime: lambda v: int(v.timestamp()),
        }


class Forward(Element):
    """
    指示合并转发信息

    nodeList (List[ForwardNode]): 转发的消息节点
    """

    type = "Forward"
    nodeList: List[ForwardNode]

    def asDisplay(self) -> str:
        return f"[合并转发:共{len(self.nodeList)}条]"


class File(Element):
    type = "File"
    id: str
    name: str
    size: int

    def asDisplay(self) -> str:
        return f"[文件:{self.name}]"


class ImageType(Enum):
    Friend = "Friend"
    Group = "Group"
    Temp = "Temp"
    Unknown = "Unknown"


image_upload_method_type_map = {
    UploadMethod.Friend: ImageType.Friend,
    UploadMethod.Group: ImageType.Group,
    UploadMethod.Temp: ImageType.Temp,
}


class Image(Element):
    type = "Image"
    ready: bool = True
    imageId: Optional[str] = None
    url: Optional[str] = None
    path: Optional[Path] = None
    base64: Optional[str] = None
    data_bytes: Optional[bytes] = None

    def __init__(
        self,
        imageId: Optional[str] = None,
        url: Optional[str] = None,
        path: Optional[Path] = None,
        base64: Optional[str] = None,
        *,
        data_bytes: Optional[bytes] = None,
        flash: Optional["FlashImage"] = None,
        **kwargs,
    ) -> None:
        if flash:
            super().__init__(**flash.dict() | {"type": "Image"})
            return
        data = {}
        data["imageId"] = imageId
        data["url"] = url
        data["path"] = path
        if data_bytes and base64:
            raise ValueError("Can't present base64 and bytes data at the same time!")
        if base64:
            data["base64"] = base64
        if data_bytes:
            data["base64"] = b64encode(data_bytes)
        super().__init__(**data, **kwargs)

    def asDisplay(self) -> str:
        return "[图片]"


class FlashImage(Image):
    type = "FlashImage"

    def __init__(
        self,
        imageId: Optional[str] = None,
        url: Optional[str] = None,
        path: Optional[Path] = None,
        base64: Optional[str] = None,
        *,
        data_bytes: Optional[bytes] = None,
        normal: Optional["Image"] = None,
        **kwargs,
    ) -> None:
        if normal:
            super().__init__(**normal.dict() | {"type": "FlashImage"})
        data = {}
        data["type"] = type
        data["imageId"] = imageId
        data["url"] = url
        data["path"] = path
        if data_bytes and base64:
            raise ValueError("Can't present base64 and bytes data at the same time!")
        if base64:
            data["base64"] = base64
        if data_bytes:
            data["base64"] = b64encode(data_bytes)
        super().__init__(**data, **kwargs)

    def asDisplay(self):
        return "[闪照]"


class Voice(Element):
    type = "Voice"
    voiceId: Optional[str]
    url: Optional[str]
    path: Optional[Path]
    base64: Optional[str]
    length: Optional[int]
    data_bytes: Optional[bytes]

    def __init__(
        self,
        voiceId: Optional[str] = None,
        url: Optional[str] = None,
        path: Optional[Path] = None,
        base64: Optional[str] = None,
        *,
        data_bytes: Optional[bytes] = None,
        **kwargs,
    ) -> None:
        data = {}
        data["type"] = type
        data["voiceId"] = voiceId
        data["url"] = url
        data["path"] = path
        if data_bytes and base64:
            raise ValueError("Can't present base64 and bytes data at the same time!")
        if base64:
            data["base64"] = base64
        if data_bytes:
            data["base64"] = b64encode(data_bytes)
        super().__init__(**data, **kwargs)

    def asDisplay(self) -> str:
        return "[语音]"


def _update_forward_refs():
    """
    Inner function.
    Update the forward references.
    """
    from .chain import MessageChain

    Quote.update_forward_refs(MessageChain=MessageChain)
    ForwardNode.update_forward_refs(MessageChain=MessageChain)



import functools
from typing import Callable, ContextManager, TypeVar, Union

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from typing_extensions import ParamSpec

from graia.ariadne.context import enter_context

P = ParamSpec("P")
R = TypeVar("R")

from graia.ariadne.exception import (
    AccountMuted,
    AccountNotFound,
    InvalidArgument,
    InvalidSession,
    InvalidVerifyKey,
    MessageTooLong,
    UnknownTarget,
    UnVerifyedSession,
)

code_exceptions_mapping = {
    1: InvalidVerifyKey,
    2: AccountNotFound,
    3: InvalidSession,
    4: UnVerifyedSession,
    5: UnknownTarget,
    6: FileNotFoundError,
    10: PermissionError,
    20: AccountMuted,
    30: MessageTooLong,
    400: InvalidArgument,
}


def validate_response(code: Union[dict, int]):
    if isinstance(code, dict):
        code = code.get("code")
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            raise exception_code
    elif isinstance(code, int):
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            raise exception_code


class ApplicationMiddlewareDispatcher(BaseDispatcher):
    always = True
    context: ContextManager

    def __init__(self, app) -> None:
        self.app = app

    def beforeExecution(self, interface: "DispatcherInterface"):
        self.context = enter_context(self.app, interface.event)
        self.context.__enter__()

    def afterExecution(self, interface: "DispatcherInterface", exception, tb):
        self.context.__exit__(exception.__class__ if exception else None, exception, tb)

    async def catch(self, interface: "DispatcherInterface"):
        from graia.ariadne.app import Ariadne

        if interface.annotation is Ariadne:
            return self.app


def app_ctx_manager(func: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(func)
    async def wrapper(self, *args: P.args, **kwargs: P.kwargs):
        with enter_context(app=self):
            return await func(self, *args, **kwargs)

    return wrapper



from graia.broadcast import Dispatchable
from graia.broadcast.entities.dispatcher import BaseDispatcher
from pydantic import BaseModel, validator

from graia.ariadne.exception import InvalidEventTypeDefinition


class MiraiEvent(Dispatchable, BaseModel):
    type: str

    @validator("type", allow_reuse=True)
    def type_limit(cls, v):
        if cls.type != v:
            raise InvalidEventTypeDefinition(
                "{0}'s type must be '{1}', not '{2}'".format(cls.__name__, cls.type, v)
            )
        return v

    class Config:
        extra = "ignore"

    class Dispatcher:
        pass
